# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np

CCI_DATA_STORE_ID = 'esa-cci'
CDC_SHORT_DATA_STORE_ID = 'esa-cdc'
CDC_LONG_DATA_STORE_ID = 'esa-climate-data-centre'
KC_CCI_DATA_STORE_ID = 'esa-cci-kc'
KC_CDC_DATA_STORE_ID = 'esa-cdc-kc'
ZARR_CCI_DATA_STORE_ID = 'esa-cci-zarr'
ZARR_CDC_DATA_STORE_ID = 'esa-cdc-zarr'

DATA_STORE_ID = 'cciodp'
KC_DATA_STORE_ID = 'ccikc'
ZARR_DATA_STORE_ID = 'ccizarr'

DATAFRAME_OPENER_ID = f'dataframe:geojson:{CDC_SHORT_DATA_STORE_ID}'
CDC_DATASET_OPENER_ID = f'dataset:zarr:{CDC_SHORT_DATA_STORE_ID}'
DATASET_OPENER_ID = f'dataset:zarr:{DATA_STORE_ID}'
VECTORDATACUBE_OPENER_ID = f'vectordatacube::{CDC_SHORT_DATA_STORE_ID}'
DATA_ARRAY_NAME = 'var_data'
OPENSEARCH_CEDA_URL = 'https://archive.opensearch.ceda.ac.uk/opensearch/request'
CCI_ODD_URL = 'https://archive.opensearch.ceda.ac.uk/' \
              'opensearch/description.xml?parentIdentifier=cci'

# to test with opensearch test, use the following two lines
# instead of the upper two
# OPENSEARCH_CEDA_URL = "http://opensearch-test.ceda.ac.uk/opensearch/request"
# CCI_ODD_URL = 'http://opensearch-test.ceda.ac.uk/' \
#               'opensearch/description.xml?parentIdentifier=cci'

DEFAULT_TILE_SIZE = 1000
DEFAULT_RETRY_BACKOFF_MAX = 40  # milliseconds
DEFAULT_RETRY_BACKOFF_BASE = 1.001
DEFAULT_NUM_RETRIES = 200

CCI_MAX_IMAGE_SIZE = 2500

COMMON_SPATIAL_COORD_VAR_NAMES = [
    'lat', 'lon', 'latitude', 'longitude',
    'latitude_centers', 'x', 'y', 'xc', 'yc'
]
COMMON_TIME_COORD_VAR_NAMES = ['time', 't', 'start_time', 'end_time']
COMMON_COORD_VAR_NAMES = COMMON_SPATIAL_COORD_VAR_NAMES + COMMON_TIME_COORD_VAR_NAMES

MONTHS = ['JANUARY', 'FEBRUARY', 'MARCH', 'APRIL', 'MAY', 'JUNE', 'JULY',
          'AUGUST', 'SEPTEMBER', 'OCTOBER', 'NOVEMBER', 'DECEMBER']

TIFF_VARS = {
    "CHANGE-CDET": {
        "year_of_change": {
            "band_index": 0,
            "description": "The year of change in the 5-year span of the two years "
                           "from which the LCC product is calulated.",
            "data_type": np.float32,
            "min_value": 1990,
            "max_value": 2019,
            "fill_value": np.nan
        },
        "probability_of_change": {
            "band_index": 1,
            "description": "Probability of the change as returned by the change "
                           "detection algorithm. Original probability values in the "
                           "range [0,1] are mapped to 0-100 intgere values.",
            "data_type": np.float32,
            "min_value": 0,
            "max_value": 100,
            "fill_value": np.nan
        },
        "reliability": {
            "band_index": 2,
            "description": "Distance between the couple of years for which the change "
                           "has been calculated",
            "data_type": np.float32,
            "min_value": 0,
            "max_value": 5,
            "fill_value": np.nan
        },
        "pcc_priority_changes": {
            "band_index": 3,
            "description": "From Post Classification Comparison of HRLC30 maps, "
                           "the class cover transition of the change in its degree of "
                           "priority. 0=no change, 1=low priority change, "
                           "2=high priority change.",
            "data_type": np.float32,
            "min_value": 0,
            "max_value": 2,
        }
    },
    "MAP-CL01": {
        "CL01": {
            "band_index": 0,
            "description": "First class as returned by the classification algorithm "
                           "according to posterior probability ranking.",
            "data_type": np.byte,
            "min_value": 0,
            "max_value": 150,
        }
    },
    "UNCERT-CL02": {
        "CL02": {
            "band_index": 0,
            "description": "Second class as returned by the classification algorithm "
                           "according to posterior probability ranking.",
            "data_type": np.byte,
            "min_value": 0,
            "max_value": 150,
        }
    },
    "UNCERT-PS01": {
        "PS01": {
            "band_index": 0,
            "description": "Posterior probability associated to the first ranked "
                           "class. Original probability values in the range [0,1] "
                           "are mapped to 0-100 integer values.",
            "data_type": np.byte,
            "min_value": 0,
            "max_value": 100,
        }
    },
    "UNCERT-PS02": {
        "PS02": {
            "band_index": 0,
            "description": "Posterior probability associated to the second ranked "
                           "class. Original probability values in the range [0,1] "
                           "are mapped to 0-100 integer values.",
            "data_type": np.byte,
            "min_value": 0,
            "max_value": 100,
        }
    },
    "UNCERT-IQIX": {
        "IQIX": {
            "band_index": 0,
            "description": "Optical imagery input quality evaluated before "
                           "classification by considering the number of valid images "
                           "at pixel level used for classification. Ranges from 0–low "
                           "quality to 3–high quality",
            "data_type": np.byte,
            "min_value": 0,
            "max_value": 3,
        }
    },
    "CL": {
        "CL": {}
    },
    "JD": {
        "JD": {}
    },
    "LC": {
        "LC": {}
    }
}

TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S"



from unittest import TestCase

from xcube_cci.normalize import normalize_dims_description
from xcube_cci.normalize import normalize_variable_dims_description


class TestNormalize(TestCase):

    def test_normalize_dims(self):
        dims_1 = dict(lat=1, lon=2, time=4)
        self.assertEqual(dims_1, normalize_dims_description(dims_1))

        dims_2 = dict(lat=1, lon=2, time=4, fgzrh=5, dfsraxt=6)
        self.assertEqual(dims_2, normalize_dims_description(dims_2))

        dims_3 = dict(latitude=1, lon=2, time=4)
        self.assertEqual(dims_1, normalize_dims_description(dims_3))

        dims_4 = dict(lat=1, longitude=2, time=4)
        self.assertEqual(dims_1, normalize_dims_description(dims_4))

        dims_5 = dict(latitude_centers=1, time=4)
        self.assertEqual(dims_1, normalize_dims_description(dims_5))

        dims_6 = dict(latitude=1, dhft=2, time=4)
        self.assertEqual(dict(lat=1, dhft=2, time=4),
                         normalize_dims_description(dims_6))

    def test_normalize_variable_dims_description(self):
        dims_1 = ['time', 'lat', 'lon']
        self.assertEqual(dims_1, normalize_variable_dims_description(dims_1))

        dims_2 =['lat', 'lon']
        self.assertEqual(dims_1, normalize_variable_dims_description(dims_2))

        dims_3 = ['latitude', 'longitude']
        self.assertEqual(dims_1, normalize_variable_dims_description(dims_3))

        dims_4 = ['latitude_centers']
        self.assertEqual(dims_1, normalize_variable_dims_description(dims_4))

        dims_5 = ['lat', 'lon', 'draeftgyhesj']
        self.assertEqual(['time', 'draeftgyhesj', 'lat', 'lon'],
                         normalize_variable_dims_description(dims_5))

        dims_6 = ['latitude_centers', 'draeftgyhesj']
        self.assertEqual(['time', 'draeftgyhesj', 'lat', 'lon'],
                         normalize_variable_dims_description(dims_6))

        dims_7 = ['lat', 'gyfdvtz', 'time']
        self.assertEqual(['lat', 'gyfdvtz', 'time'],
                         normalize_variable_dims_description(dims_7))

        dims_8 = ['gyfdvtz']
        self.assertEqual(['gyfdvtz'],
                         normalize_variable_dims_description(dims_8))



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import bisect
import copy
import itertools
import json
import logging
import math
import numcodecs
import time
import warnings
from abc import abstractmethod, ABCMeta
from collections.abc import MutableMapping
from numcodecs import Blosc
from typing import Iterator, Any, List, Dict, Tuple, Callable, Iterable, \
    KeysView, Mapping

import numpy as np
import pandas as pd

from .cciodp import CciOdp
from .constants import COMMON_COORD_VAR_NAMES
from .constants import TIMESTAMP_FORMAT
from .timerangegetter import extract_time_range_as_strings
from .timerangegetter import TimeRangeGetter

_MIN_CHUNK_SIZE = 512 * 512
_MAX_CHUNK_SIZE = 2048 * 2048

_STATIC_ARRAY_COMPRESSOR_PARAMS = dict(
    cname='zstd', clevel=1, shuffle=Blosc.SHUFFLE, blocksize=0
)
_STATIC_ARRAY_COMPRESSOR_CONFIG = dict(
    id='blosc', **_STATIC_ARRAY_COMPRESSOR_PARAMS
)
_STATIC_ARRAY_COMPRESSOR = Blosc(**_STATIC_ARRAY_COMPRESSOR_PARAMS)

_LOG = logging.getLogger()


def _dict_to_bytes(d: Dict):
    return _str_to_bytes(json.dumps(d, indent=2))


def _str_to_bytes(s: str):
    return bytes(s, encoding='utf-8')


class RemoteChunkStore(MutableMapping, metaclass=ABCMeta):
    """
    A remote Zarr Store.

    :param data_id: The identifier of the data resource
    :param cube_params: A mapping containing additional parameters to define
        the data set.
    :param observer: An optional callback function called when remote requests
        are mode: observer(**kwargs).
    :param trace_store_calls: Whether store calls shall be printed
        (for debugging).
    """

    def __init__(self,
                 data_id: str,
                 cube_params: Mapping[str, Any] = None,
                 observer: Callable = None,
                 trace_store_calls=False):
        if not cube_params:
            cube_params = {}
        self._variable_names = cube_params.get('variable_names',
                                               self.get_all_variable_names())
        self._attrs = {}
        self._observers = [observer] if observer is not None else []
        self._trace_store_calls = trace_store_calls

        self._dataset_name = data_id
        self._time_ranges = self.get_time_ranges(data_id, cube_params)
        is_climatology = 'AEROSOL.climatology' in data_id
        logging.debug('Determined time ranges')
        if not self._time_ranges:
            raise ValueError('Could not determine any valid time stamps')
        self._time_chunking = self.get_time_chunking()
        if isinstance(self._time_chunking, List):
            self._time_chunking = self._time_chunking[0]
        if self._time_chunking > 1 and 'time_range' in cube_params:
            all_ranges = self.get_time_ranges(data_id, {})
            first_index = all_ranges.index(self._time_ranges[0])
            new_first_index = math.floor(
                first_index / self._time_chunking
            ) * self._time_chunking
            last_index = all_ranges.index(self._time_ranges[-1])
            new_last_index = max(new_first_index + 1,
                                 math.ceil(last_index / self._time_chunking)
                                 * self._time_chunking)
            self._time_ranges = all_ranges[new_first_index:new_last_index]

        if is_climatology:
            t_array = np.array(range(1, 13), dtype=np.int8)
            t_bnds_array = None
            time_coverage_start = None
            time_coverage_end = None
        else:
            t_array = [s.to_pydatetime()
                       + 0.5 * (e.to_pydatetime() - s.to_pydatetime())
                       for s, e in self._time_ranges]
            t_array = np.array(t_array).astype('datetime64[s]').astype(np.int64)
            t_bnds_array = np.array(self._time_ranges).\
                astype('datetime64[s]').astype(np.int64)
            time_coverage_start = self._time_ranges[0][0]
            time_coverage_end = self._time_ranges[-1][1]
            cube_params['time_range'] = (extract_time_range_as_strings(
                cube_params.get(
                    'time_range', self.get_default_time_range(data_id)
                )
            ))

        self._vfs = {}
        self._var_name_to_ranges = {}
        self._ranges_to_indexes = {}
        self._ranges_to_var_names = {}

        bbox = cube_params.get('bbox', None)
        lon_size = -1
        lat_size = -1
        self._dimension_chunk_offsets = {}
        self._dimensions = self.get_dimensions()
        self._num_data_var_chunks_not_in_vfs = 0

        coords_data = self.get_coords_data(data_id)

        logging.debug('Determined coordinates')

        if is_climatology:
            coords_data['month'] = {}
            coords_data['month']['size'] = len(t_array)
            coords_data['month']['data'] = t_array
        else:
            coords_data['time'] = {}
            coords_data['time']['size'] = len(t_array)
            coords_data['time']['data'] = t_array
            if 'time_bounds' in coords_data:
                coords_data.pop('time_bounds')
            coords_data['time_bnds'] = {}
            coords_data['time_bnds']['size'] = len(t_bnds_array)
            coords_data['time_bnds']['data'] = t_bnds_array

        sorted_coords_names = list(coords_data.keys())
        sorted_coords_names.sort()

        lat_min_offset = -1
        lat_max_offset = -1
        lon_min_offset = -1
        lon_max_offset = -1

        time_coord_names = ["time", "time_bnds", "month", "t", "start_time", "end_time"]
        for coord_name in sorted_coords_names:
            if coord_name in time_coord_names:
                continue
            coord_attrs = self.get_attrs(coord_name)
            coord_attrs['_ARRAY_DIMENSIONS'] = coord_attrs['dimensions']
            coord_data = coords_data[coord_name]['data']
            if bbox is not None and \
                    (coord_name == 'lat' or coord_name == 'latitude'
                     or coord_name == 'y'):
                if coord_data[0] < coord_data[-1]:
                    lat_min_offset = bisect.bisect_left(coord_data, bbox[1])
                    lat_max_offset = bisect.bisect_right(coord_data, bbox[3])
                else:
                    lat_min_offset = \
                        len(coord_data) - bisect.bisect_left(
                            coord_data[::-1], bbox[3]
                        )
                    lat_max_offset = \
                        len(coord_data) - bisect.bisect_right(
                            coord_data[::-1], bbox[1]
                        )
                coords_data = self._adjust_coord_data(coord_name,
                                                      lat_min_offset,
                                                      lat_max_offset,
                                                      coords_data,
                                                      coord_attrs)
                coord_data = coords_data[coord_name]['data']
            elif bbox is not None and \
                    (coord_name == 'lon' or coord_name == 'longitude'
                     or coord_name == 'x'):
                lon_min_offset = bisect.bisect_left(coord_data, bbox[0])
                lon_max_offset = bisect.bisect_right(coord_data, bbox[2])
                coords_data = self._adjust_coord_data(coord_name,
                                                      lon_min_offset,
                                                      lon_max_offset,
                                                      coords_data,
                                                      coord_attrs)
                coord_data = coords_data[coord_name]['data']
            elif bbox is not None and (
                    coord_name == 'latitude_bounds' or
                    coord_name == 'lat_bounds' or
                    coord_name == 'latitude_bnds' or
                    coord_name == 'lat_bnds'
            ):
                coords_data = self._adjust_coord_data(coord_name,
                                                      lat_min_offset,
                                                      lat_max_offset,
                                                      coords_data,
                                                      coord_attrs)
                coord_data = coords_data[coord_name]['data']
            elif bbox is not None and (
                    coord_name == 'longitude_bounds' or
                    coord_name == 'lon_bounds' or
                    coord_name == 'longitude_bnds' or
                    coord_name == 'lon_bnds'
            ):
                coords_data = self._adjust_coord_data(coord_name,
                                                      lon_min_offset,
                                                      lon_max_offset,
                                                      coords_data,
                                                      coord_attrs)
                coord_data = coords_data[coord_name]['data']
            if len(coord_data) > 0:
                coord_array = np.array(coord_data)
                self._add_static_array(coord_name, coord_array, coord_attrs)
            else:
                shape = coords_data[coord_name].get(
                    'shape', coords_data[coord_name].get('size')
                )
                if isinstance(shape, tuple):
                    shape = list(shape)
                else:
                    shape = [shape]
                chunk_size = coords_data[coord_name]['chunkSize']
                if not isinstance(chunk_size, List):
                    chunk_size = [chunk_size]
                encoding = self.get_encoding(coord_name)
                self._add_remote_array(coord_name, shape, chunk_size,
                                       encoding, coord_attrs)

        time_dim_name = "time"
        if is_climatology:
            time_dim_name = 'month'
        if "Time" in self._dimensions:
            time_dim_name = "Time"
        elif "nbmonth" in self._dimensions:
            time_dim_name = "nbmonth"
        time_bounds_name = f"{time_dim_name}_bnds"
        if is_climatology:
            month_attrs = {
                "_ARRAY_DIMENSIONS": [time_dim_name],
                "standard_name": "month"
            }
            self._add_static_array('month', t_array, month_attrs)
        else:
            time_attrs = {
                "_ARRAY_DIMENSIONS": [time_dim_name],
                "units": "seconds since 1970-01-01T00:00:00Z",
                "calendar": "proleptic_gregorian",
                "standard_name": time_dim_name,
                "bounds": time_bounds_name,
            }
            time_bnds_attrs = {
                "_ARRAY_DIMENSIONS": [time_dim_name, 'bnds'],
                "units": "seconds since 1970-01-01T00:00:00Z",
                "calendar": "proleptic_gregorian",
                "standard_name": time_bounds_name,
            }
            self._add_static_array(time_dim_name, t_array, time_attrs)
            self._add_static_array(
                time_bounds_name, t_bnds_array, time_bnds_attrs
            )

        coordinate_names = [coord for coord in coords_data.keys()
                            if coord not in COMMON_COORD_VAR_NAMES]
        if time_dim_name not in coordinate_names:
            coordinate_names.append(time_dim_name)
        if time_bounds_name not in coordinate_names:
            coordinate_names.append(time_bounds_name)
        coordinate_names = ' '.join(coordinate_names)

        global_attrs = dict(
            Conventions='CF-1.7',
            coordinates=coordinate_names,
            title=data_id,
            date_created=pd.Timestamp.now().isoformat(),
            processing_level=self._dataset_name.split('.')[3]
        )
        if not is_climatology:
            global_attrs['time_coverage_start'] = \
                time_coverage_start.isoformat()
            global_attrs['time_coverage_end'] = \
                time_coverage_end.isoformat()
            global_attrs['time_coverage_duration'] = \
                (time_coverage_end - time_coverage_start).isoformat()

        self._time_indexes = {}
        for coord_name in sorted_coords_names:
            self._time_indexes[coord_name] = -1
            if coord_name == time_dim_name or coord_name == time_bounds_name:
                self._time_indexes[coord_name] = 0
        remove = []
        logging.debug('Adding variables to dataset ...')
        for variable_name in self._variable_names:
            var_attrs = self.get_attrs(variable_name)
            grid_mapping_name = var_attrs.get('grid_mapping', variable_name)
            if grid_mapping_name not in self._variable_names:
                self._variable_names.append(grid_mapping_name)
        for variable_name in self._variable_names:
            if variable_name in coords_data or variable_name == time_bounds_name:
                remove.append(variable_name)
                continue
            var_encoding = self.get_encoding(variable_name)
            var_attrs = self.get_attrs(variable_name)
            dimensions = var_attrs.get('dimensions', [])
            self._maybe_adjust_attrs(lon_size, lat_size, var_attrs)
            chunk_sizes = var_attrs.get('chunk_sizes', [-1] * len(dimensions))
            if isinstance(chunk_sizes, int):
                chunk_sizes = [chunk_sizes]
            if len(dimensions) > 0 and time_dim_name not in dimensions \
                    and self._optimise_chunking():
                dimensions.insert(0, time_dim_name)
                chunk_sizes.insert(0, 1)
            var_attrs.update(_ARRAY_DIMENSIONS=dimensions)
            sizes = []
            self._time_indexes[variable_name] = -1
            time_dimension = -1
            for i, coord_name in enumerate(dimensions):
                if coord_name in coords_data:
                    sizes.append(coords_data[coord_name]['size'])
                else:
                    sizes.append(self._dimensions.get(coord_name))
                if coord_name == time_dim_name:
                    self._time_indexes[variable_name] = i
                    time_dimension = i
                    chunk_sizes[i] = self._time_chunking
                if chunk_sizes[i] == -1:
                    chunk_sizes[i] = sizes[i]
            if len(sizes) == 0:
                chunk_sizes = []
            var_attrs['shape'] = sizes
            var_attrs['size'] = math.prod(sizes)
            if var_encoding.get('dtype', '') == 'bytes1024':
                if 'grid_mapping_name' in var_attrs:
                    var_encoding['dtype'] = 'U'
                elif len(dimensions) == 1 and sizes[0] < 512 * 512:
                    _LOG.info(f"Variable '{variable_name}' is encoded as "
                              f"string. Will convert it to metadata.")
                    variable = {variable_name: sizes[0]}
                    var_data = self.get_variable_data(data_id, variable)
                    global_attrs[variable_name] = \
                        [var.decode('utf-8')
                         for var in var_data[variable_name]['data']]
                    remove.append(variable_name)
                    continue
                else:
                    warnings.warn(f"Variable '{variable_name}' is encoded as "
                                  f"string. Will omit it from the dataset.")
                    remove.append(variable_name)
                    continue
            if self._optimise_chunking():
                chunk_sizes = self._adjust_chunk_sizes(chunk_sizes,
                                                       sizes,
                                                       time_dimension)
            var_attrs['chunk_sizes'] = chunk_sizes
            if len(var_attrs.get('file_dimensions', [])) < len(dimensions):
                var_attrs['file_chunk_sizes'] = chunk_sizes[1:].copy()
            elif len(var_attrs.get('file_dimensions', [])) > 0:
                var_attrs['file_chunk_sizes'] = chunk_sizes.copy()
                if time_dimension >= 0:
                    var_attrs['file_chunk_sizes'][time_dimension] = self._time_chunking
            self._add_remote_array(variable_name,
                                   sizes,
                                   chunk_sizes,
                                   var_encoding,
                                   var_attrs)
        logging.debug(f"Added a total of {len(self._variable_names)} variables "
                      f"to the data set")
        for r in remove:
            self._variable_names.remove(r)
        cube_params['variable_names'] = self._variable_names
        global_attrs['history'] = [dict(
            program=f'{self._class_name}',
            cube_params=cube_params
        )]
        # setup Virtual File System (vfs)
        self._vfs['.zgroup'] = _dict_to_bytes(dict(zarr_format=2))
        self._vfs['.zattrs'] = _dict_to_bytes(global_attrs)

    def _adjust_coord_data(self, coord_name: str, min_offset: int,
                           max_offset: int, coords_data, dim_attrs: dict):
        self._dimension_chunk_offsets[coord_name] = min_offset
        if min_offset == max_offset:
            if min_offset == len(coords_data[coord_name]['data']) - 1:
                min_offset -= 1
            else:
                max_offset += 1
        coord_data = coords_data[coord_name]['data'][min_offset:max_offset]
        shape = coord_data.shape
        self._set_chunk_sizes(dim_attrs, shape, 'chunk_sizes')
        self._set_chunk_sizes(dim_attrs, shape, 'file_chunk_sizes')
        dim_attrs['size'] = coord_data.size
        if 'shape' in dim_attrs:
            dim_attrs['shape'] = list(shape)
        if len(shape) == 1:
            self._dimensions[coord_name] = coord_data.size
        coords_data[coord_name]['size'] = coord_data.size
        coords_data[coord_name]['chunkSize'] = dim_attrs['chunk_sizes']
        coords_data[coord_name]['data'] = coord_data
        return coords_data

    @staticmethod
    def _set_chunk_sizes(dim_attrs, shape, name):
        chunk_sizes = dim_attrs.get(name, 1000000)
        if isinstance(chunk_sizes, int):
            dim_attrs[name] = min(chunk_sizes, shape[0])
        else:
            # chunk sizes is list of ints
            for i, chunk_size in enumerate(chunk_sizes):
                chunk_sizes[i] = min(chunk_size, shape[i])
            dim_attrs[name] = chunk_sizes

    @classmethod
    def _maybe_adjust_attrs(cls, lon_size, lat_size, var_attrs):
        cls._maybe_adjust_to('lat', 'latitude', lat_size, var_attrs)
        cls._maybe_adjust_to('lon', 'longitude', lon_size, var_attrs)

    @classmethod
    def _maybe_adjust_to(
            cls, first_name, second_name, adjusted_size, var_attrs
    ):
        if adjusted_size == -1:
            return
        try:
            index = var_attrs['dimensions'].index(first_name)
        except ValueError:
            try:
                index = var_attrs['dimensions'].index(second_name)
            except ValueError:
                index = -1
        if index > 0:
            var_attrs['shape'][index] = adjusted_size
            if 'chunk_sizes' in var_attrs:
                var_attrs['chunk_sizes'][index] = \
                    min(var_attrs['chunk_sizes'][index], adjusted_size)
                var_attrs['file_chunk_sizes'][index] = \
                    min(var_attrs['file_chunk_sizes'][index], adjusted_size)

    @classmethod
    def _adjust_chunk_sizes(cls, chunks, sizes, time_dimension):
        # check if we can read in everything as chunks only chunked by time
        sum_sizes = np.prod(sizes, dtype=np.int64)
        if time_dimension >= 0:
            sum_sizes = sum_sizes / \
                        sizes[time_dimension] * chunks[time_dimension]
        if sum_sizes < _MAX_CHUNK_SIZE:
            best_chunks = sizes.copy()
            if time_dimension >= 0:
                best_chunks[time_dimension] = chunks[time_dimension]
            return best_chunks
        # check whether default chunks are acceptable:
        sum_chunks = np.prod(chunks, dtype=np.int64)
        if cls._is_of_acceptable_chunk_size(sum_chunks):
            chunks_are_acceptable = True
            for i in range(len(chunks)):
                if i != time_dimension and (chunks[i] > sizes[i] or sizes[i] % chunks[i] > 0):
                    chunks_are_acceptable = False
                    break
            if chunks_are_acceptable:
                return chunks
        # determine valid values for a chunk size.
        # A value is valid if the size can be divided by it without remainder
        valid_chunk_sizes = []
        for i, chunk, size in zip(range(len(chunks)), chunks, sizes):
            # do not rechunk time dimension
            if i == time_dimension:
                valid_chunk_sizes.append([chunk])
                continue
            valid_dim_chunk_sizes = common_divisors(size)
            valid_chunk_sizes.append(valid_dim_chunk_sizes)
        # recursively determine the chunking most similar to the original
        # file chunking with an acceptable size
        initial_best_chunks = [vcs[-1] for vcs in valid_chunk_sizes]
        initial_test_chunks = [vcs[0] for vcs in valid_chunk_sizes]
        best_deviation = 0
        chunks, chunk_size, chunk_deviation = cls._get_best_chunks(
            initial_test_chunks, valid_chunk_sizes, initial_best_chunks,
            0, 0, time_dimension, best_deviation
        )
        if not cls._is_of_acceptable_chunk_size(np.prod(chunks, dtype=np.int64)):
            # return original chunks
            return [vcs[0] for vcs in valid_chunk_sizes]
        return chunks

    @classmethod
    def _get_best_chunks(cls, chunks, valid_sizes, best_chunks,
                         best_chunk_size, index, time_dimension,
                         best_deviation):
        for valid_size in valid_sizes[index]:
            test_chunks = chunks.copy()
            test_chunks[index] = valid_size
            if index < len(chunks) - 1:
                test_chunks, test_chunk_size, test_deviation = \
                    cls._get_best_chunks(test_chunks, valid_sizes, best_chunks,
                                         best_chunk_size, index + 1,
                                         time_dimension, best_deviation)
                if test_chunks == best_chunks:
                    continue
            else:
                test_chunk_size = np.prod(test_chunks, dtype=np.int64)
                test_deviation = cls.determine_deviation(test_chunks, time_dimension, test_chunk_size)
            if cls._is_of_acceptable_chunk_size(test_chunk_size):
                if test_deviation > best_deviation:
                    best_chunk_size = test_chunk_size
                    best_chunks = test_chunks.copy()
                    best_deviation = test_deviation
                elif test_deviation == best_deviation:
                    # choose the one where values are more similar
                    where = np.full(len(test_chunks), fill_value=True)
                    where[time_dimension] = False
                    test_min_chunk = np.max(test_chunks, initial=0,
                                            where=where)
                    best_min_chunk = np.max(best_chunks, initial=0,
                                            where=where)
                    if best_min_chunk < test_min_chunk:
                        best_chunk_size = test_chunk_size
                        best_chunks = test_chunks.copy()
                        best_deviation = test_deviation
        return best_chunks, best_chunk_size, best_deviation

    @classmethod
    def index_of_list(cls, lists, indexes):
        index_list = []
        for i, index in enumerate(indexes):
            index_list.append(lists[i].index(index))
        return index_list

    @classmethod
    def compare_lists(cls, list1, list2):
        deviation = 0
        for i in range(len(list1)):
            deviation += abs(list1[i] - list2[i])
        return deviation

    @classmethod
    def determine_deviation(cls, list1, time_dimension, chunk_size):
        deviation = 0
        for i in range(0, len(list1)):
            if i == time_dimension:
                continue
            for j in range(i+1, len(list1)):
                if j == time_dimension:
                    continue
                deviation += abs((list1[i] / list1[j]) - 1)
        return chunk_size / pow(1 + deviation, 2)

    @classmethod
    def _is_of_acceptable_chunk_size(cls, size: int):
        return _MIN_CHUNK_SIZE <= size <= _MAX_CHUNK_SIZE

    @abstractmethod
    def _optimise_chunking(self) -> bool:
        pass

    @abstractmethod
    def get_time_ranges(
            self, cube_id: str, cube_params: Mapping[str, Any]
    ) -> List[Tuple]:
        pass

    @abstractmethod
    def get_time_chunking(self) -> int:
        pass

    @abstractmethod
    def get_default_time_range(self, ds_id: str) -> Tuple[str, str]:
        return '', ''

    @abstractmethod
    def get_all_variable_names(self) -> List[str]:
        pass

    @abstractmethod
    def get_dimensions(self) -> Dict[str, int]:
        pass

    @abstractmethod
    def get_coords_data(self, dataset_id: str) -> dict:
        pass

    @abstractmethod
    def get_variable_data(
            self, dataset_id: str, variable_names: Dict[str, int]
    ):
        pass

    def add_observer(self, observer: Callable):
        """
        Add a request observer.

        :param observer: A callback function called when remote requests are
            mode: observer(**kwargs).
        """
        self._observers.append(observer)

    @abstractmethod
    def get_encoding(self, band_name: str) -> Dict[str, Any]:
        """
        Get the encoding settings for band (variable) *band_name*.
        Must at least contain "dtype" whose value is a numpy array-protocol
        type string. Refer to
        https://docs.scipy.org/doc/numpy/reference/arrays.interface.html#arrays-interface
        and zarr format 2 spec.
        """

    @abstractmethod
    def get_attrs(self, band_name: str) -> Dict[str, Any]:
        """
        Get any metadata attributes for band (variable) *band_name*.
        """

    def request_time_range(self, time_index: int) -> Tuple:
        start_index = time_index * self._time_chunking
        end_index = ((time_index + 1) * self._time_chunking) - 1
        start_time = self._time_ranges[start_index][0]
        end_time = self._time_ranges[end_index][1]
        return start_time, end_time

    def _add_static_array(self, name: str, array: np.ndarray, attrs: Dict):
        shape = list(map(int, array.shape))
        dtype = str(array.dtype.str)
        order = "C"
        array_metadata = {
            "zarr_format": 2,
            "chunks": shape,
            "shape": shape,
            "dtype": dtype,
            "fill_value": None,
            "compressor": _STATIC_ARRAY_COMPRESSOR_CONFIG,
            "filters": None,
            "order": order,
        }
        chunk_key = '.'.join(['0'] * array.ndim)
        self._vfs[name] = _str_to_bytes('')
        self._vfs[name + '/.zarray'] = _dict_to_bytes(array_metadata)
        self._vfs[name + '/.zattrs'] = _dict_to_bytes(attrs)
        self._vfs[name + '/' + chunk_key] = \
            _STATIC_ARRAY_COMPRESSOR.encode(array.tobytes(order=order))

    def _add_remote_array(self,
                          name: str,
                          shape: List[int],
                          chunks: List[int],
                          encoding: Dict[str, Any],
                          attrs: Dict):
        array_metadata = dict(zarr_format=2,
                              shape=shape,
                              chunks=chunks,
                              compressor=None,
                              fill_value=None,
                              filters=None,
                              order='C')
        array_metadata.update(encoding)
        self._vfs[name] = _str_to_bytes('')
        self._vfs[name + '/.zarray'] = _dict_to_bytes(array_metadata)
        self._vfs[name + '/.zattrs'] = _dict_to_bytes(attrs)
        nums = []
        for i in range(len(shape)):
            nums.append(math.ceil(shape[i] / chunks[i]))
        nums = np.array(nums)
        ranges = tuple(map(range, map(int, nums)))
        self._var_name_to_ranges[name] = ranges
        if ranges not in self._ranges_to_indexes:
            self._ranges_to_indexes[ranges] = list(itertools.product(*ranges))
        if ranges not in self._ranges_to_var_names:
            self._ranges_to_var_names[ranges] = []
        self._ranges_to_var_names[ranges].append(name)
        self._num_data_var_chunks_not_in_vfs += \
            len(self._ranges_to_indexes[ranges])

    def _fetch_chunk(self, key: str, var_name: str,
                     chunk_index: Tuple[int, ...]) -> bytes:
        time_index = self._time_indexes.get(var_name, 0)
        if time_index >= 0:
            request_time_range = self.request_time_range(chunk_index[time_index])
        else:
            request_time_range = self._time_ranges[0][0], self._time_ranges[-1][1]
        t0 = time.perf_counter()
        try:
            exception = None
            chunk_data = self.fetch_chunk(key, var_name, chunk_index,
                                          request_time_range)
        except Exception as e:
            exception = e
            chunk_data = None
        duration = time.perf_counter() - t0

        for observer in self._observers:
            observer(var_name=var_name,
                     chunk_index=chunk_index,
                     # bbox=request_bbox,
                     time_range=request_time_range,
                     duration=duration,
                     exception=exception)

        if exception:
            raise exception

        return chunk_data

    @abstractmethod
    def fetch_chunk(self,
                    key: str,
                    var_name: str,
                    chunk_index: Tuple[int, ...],
                    time_range: Tuple
                    ) -> bytes:
        """
        Fetch chunk data from remote.

        :param key: The original chunk key being retrieved.
        :param var_name: Variable name
        :param chunk_index: chunk index
        :param time_range: Requested time range
        :return: chunk data as raw bytes
        """
        pass

    @property
    def _class_name(self):
        return self.__module__ + '.' + self.__class__.__name__

    ############################################################################
    # Zarr Store (MutableMapping) implementation
    ############################################################################

    def keys(self) -> KeysView[str]:
        if self._trace_store_calls:
            print(f'{self._class_name}.keys()')
        self._build_missing_vfs_entries()
        return self._vfs.keys()

    def listdir(self, key: str) -> Iterable[str]:
        if self._trace_store_calls:
            print(f'{self._class_name}.listdir(key={key!r})')
        if key == '':
            return list((k for k in self._vfs.keys() if '/' not in k))
        else:
            prefix = key + '/'
            start = len(prefix)
            return list((k for k in self._vfs.keys() if
                         k.startswith(prefix) and k.find('/', start) == -1))

    def getsize(self, key: str) -> int:
        if self._trace_store_calls:
            print(f'{self._class_name}.getsize(key={key!r})')
        return len(self._vfs[key]) + self._num_data_var_chunks_not_in_vfs

    def __iter__(self) -> Iterator[str]:
        if self._trace_store_calls:
            print(f'{self._class_name}.__iter__()')
        self._build_missing_vfs_entries()
        return iter(self._vfs.keys())

    def __len__(self) -> int:
        if self._trace_store_calls:
            print(f'{self._class_name}.__len__()')
        return len(self._vfs.keys()) + self._num_data_var_chunks_not_in_vfs

    def __contains__(self, key) -> bool:
        if self._trace_store_calls:
            print(f'{self._class_name}.__contains__(key={key!r})')
        if key in self._vfs:
            return True
        self._try_building_vfs_entry(key)
        return key in self._vfs

    def __getitem__(self, key: str) -> bytes:
        if self._trace_store_calls:
            print(f'{self._class_name}.__getitem__(key={key!r})')
        try:
            value = self._vfs[key]
        except KeyError:
            self._try_building_vfs_entry(key)
            value = self._vfs[key]
        if isinstance(value, tuple):
            return self._fetch_chunk(key, *value)
        return value

    def _try_building_vfs_entry(self, key):
        if '/' in key:
            name, chunk_index_part = key.split('/')
            try:
                chunk_indexes = \
                    tuple(int(chunk_index) for chunk_index in
                          chunk_index_part.split('.'))
            except ValueError:
                # latter part of key does not consist of chunk indexes
                return
            # build vfs entry of this chunk index
            # for all variables that have this range
            ranges = self._var_name_to_ranges[name]
            indexes = self._ranges_to_indexes[ranges]
            if chunk_indexes in indexes:
                for var_name in self._ranges_to_var_names[ranges]:
                    self._vfs[var_name + '/' + chunk_index_part] = \
                        var_name, chunk_indexes
                    self._num_data_var_chunks_not_in_vfs -= 1
                indexes.remove(chunk_indexes)

    def _build_missing_vfs_entries(self):
        for name, ranges in self._var_name_to_ranges.items():
            indexes = self._ranges_to_indexes[ranges]
            for index in indexes:
                filename = '.'.join(map(str, index))
                self._vfs[name + '/' + filename] = name, index

    def __setitem__(self, key: str, value: bytes) -> None:
        if self._trace_store_calls:
            print(
                f'{self._class_name}.__setitem__(key={key!r}, value={value!r})')
        raise TypeError(f'{self._class_name} is read-only')

    def __delitem__(self, key: str) -> None:
        if self._trace_store_calls:
            print(f'{self._class_name}.__delitem__(key={key!r})')
        raise TypeError(f'{self._class_name} is read-only')


class CciChunkStore(RemoteChunkStore):
    """
    A remote Zarr Store using the ESA CCI Climate Data Centre as backend.

    :param cci_cdc: CCI CDC instance.
    :param dataset_id: ID of the CDC dataset for which the chunkstore refers
    :param cube_params:
    :param observer: An optional callback function called when remote requests 
        are mode: observer(**kwargs).
    :param trace_store_calls: Whether store calls shall be printed
        (for debugging).
    """

    _SAMPLE_TYPE_TO_DTYPE = {
        # Note: Sentinel Hub currently only supports unsigned
        # integer values therefore requesting INT8 or INT16
        # will return the same as UINT8 or UINT16 respectively.
        'uint8': '|u1',
        'uint16': '<u2',
        'uint32': '<u4',
        'int8': '|u1',
        'int16': '<u2',
        'int32': '<u4',
        'float32': '<f4',
        'float64': '<f8',
    }

    def __init__(self,
                 cci_cdc: CciOdp,
                 dataset_id: str,
                 cube_params: Mapping[str, Any] = None,
                 observer: Callable = None,
                 trace_store_calls=False):
        self._cci_cdc = cci_cdc
        if dataset_id not in self._cci_cdc.dataset_names:
            raise ValueError(f'Data ID {dataset_id} not provided by CDC.')
        self._metadata = self._cci_cdc.get_dataset_metadata(dataset_id)
        self._time_range_getter = TimeRangeGetter(self._cci_cdc, self._metadata)
        super().__init__(dataset_id,
                         cube_params,
                         observer=observer,
                         trace_store_calls=trace_store_calls)

    def _optimise_chunking(self) -> bool:
        return self._cci_cdc.get_data_type() != "vectordatacube"

    def get_time_ranges(self, dataset_id: str,
                        cube_params: Mapping[str, Any]) -> List[Tuple]:
        return self._time_range_getter.get_time_ranges(dataset_id, cube_params)

    def get_time_chunking(self) -> int:
        return self._metadata.get(
            "time_chunking", self.get_attrs('time').get('file_chunk_sizes', 1)
        )

    def get_default_time_range(self, ds_id: str):
        return self._time_range_getter.get_default_time_range(ds_id)

    def get_all_variable_names(self) -> List[str]:
        var_names = [variable['var_id'] for variable in self._metadata['variables']]
        if len(var_names) == 0:
            var_names = [k for k, v in self._metadata.get("variable_infos", {}).items()]
        return var_names

    def get_dimensions(self) -> Dict[str, int]:
        return copy.copy(self._metadata['dimensions'])

    def get_coords_data(self, dataset_id: str):
        var_names, coord_names = self._cci_cdc.var_and_coord_names(dataset_id)
        coords_dict = {}
        for coord_name in coord_names:
            coords_dict[coord_name] = self.get_attrs(coord_name).get('size')
        dimension_data = self.get_variable_data(dataset_id, coords_dict)
        if len(dimension_data) == 0:
            # no valid data found in indicated time range,
            # let's set this broader
            dimension_data = self._cci_cdc.get_variable_data(dataset_id,
                                                             coords_dict)
        return dimension_data

    def get_variable_data(self, dataset_id: str, variable_dict: Dict[str, int]):
        try:
            start = self._time_ranges[0][0].strftime(TIMESTAMP_FORMAT)
            end = self._time_ranges[0][1].strftime(TIMESTAMP_FORMAT)
        except AttributeError:
            start = self._time_ranges[0][0]
            end = self._time_ranges[0][1]
        except ValueError:
            start = self._time_ranges[0][0]
            end = self._time_ranges[0][1]
        return self._cci_cdc.get_variable_data(dataset_id, variable_dict, start, end)

    def get_encoding(self, var_name: str) -> Dict[str, Any]:
        encoding_dict = {
            'fill_value': self.get_attrs(var_name).get('fill_value'),
            'dtype': self.get_attrs(var_name).get('data_type')}
        if var_name == "geometry" and self._cci_cdc.get_data_type() == "vectordatacube":
            json_codec = numcodecs.JSON()
            encoding_dict["filters"] = [json_codec.get_config()]
        return encoding_dict

    def get_attrs(self, var_name: str) -> Dict[str, Any]:
        if var_name not in self._attrs:
            self._attrs[var_name] = copy.deepcopy(
                self._metadata.get('variable_infos', {}).get(var_name, {}))
        return self._attrs[var_name]

    def fetch_chunk(self,
                    key: str,
                    var_name: str,
                    chunk_index: Tuple[int, ...],
                    time_range: Tuple) -> bytes:

        start_time, end_time = time_range
        identifier = self._cci_cdc.get_dataset_id(self._dataset_name)
        try:
            start_time = start_time.tz_localize(None).isoformat()
            end_time = end_time.tz_localize(None).isoformat()
        except TypeError:
            # use unconverted time range values
            pass
        dim_indexes = self._get_dimension_indexes_for_chunk(var_name,
                                                            chunk_index)
        request = dict(parentIdentifier=identifier,
                       varNames=[var_name],
                       startDate=start_time,
                       endDate=end_time,
                       drsId=self._dataset_name,
                       fileFormat='.nc'
                       )
        data = self._cci_cdc.get_data_chunk(request, dim_indexes)
        if not data:
            warnings.warn(f'{key}: cannot fetch chunk for variable '
                          f'{var_name!r} and time_range {time_range!r}.')
            expected_chunk_size, dtype_size = \
                self._determine_expected_chunk_size(var_name)
            data_type = self.get_attrs(var_name).get('data_type')
            dtype = np.dtype(self._SAMPLE_TYPE_TO_DTYPE[data_type])
            fill_value = self.get_attrs(var_name).get('fill_value')
            var_array = np.full(
                shape=int(expected_chunk_size / dtype_size),
                fill_value=fill_value,
                dtype=dtype)
            return var_array.tobytes()
        if self._time_chunking > 1 and self._cci_cdc.get_data_type() != "vectordatacube":
            expected_chunk_size, dtype_size = \
                self._determine_expected_chunk_size(var_name)
            data_length = len(data)
            if data_length < expected_chunk_size:
                var_dimensions = self.get_attrs(var_name).get('dimensions', [])
                data_type = self.get_attrs(var_name).get('data_type')
                dtype = np.dtype(self._SAMPLE_TYPE_TO_DTYPE[data_type])
                fill_value = self.get_attrs(var_name).get('fill_value')
                var_array = np.full(
                    shape=int((expected_chunk_size - data_length) / dtype_size),
                    fill_value=fill_value,
                    dtype=dtype)
                if chunk_index[var_dimensions.index('time')] == 0:
                    return var_array.tobytes() + data
                else:
                    return data + var_array.tobytes()
        _LOG.info(f'Fetched chunk for ({chunk_index})"{var_name}"')
        return data

    def _determine_expected_chunk_size(self, var_name: str):
        chunk_sizes = self.get_attrs(var_name).get('chunk_sizes', {})
        expected_chunk_size = np.prod(chunk_sizes)
        data_type = self.get_attrs(var_name).get('data_type')
        dtype = np.dtype(self._SAMPLE_TYPE_TO_DTYPE[data_type])
        return expected_chunk_size * dtype.itemsize, dtype.itemsize

    def _get_dimension_indexes_for_chunk(self, var_name: str,
                                         chunk_index: Tuple[int, ...]) -> tuple:
        dim_indexes = []
        var_dimensions = self.get_attrs(var_name).get('file_dimensions', [])
        chunk_sizes = self.get_attrs(var_name).get('file_chunk_sizes', [])
        if not isinstance(chunk_sizes, List):
            chunk_sizes = [chunk_sizes]
        offset = 0
        # dealing with the case that time has been added
        # as additional first dimension
        if len(chunk_index) > len(chunk_sizes):
            offset = 1
        for i, var_dimension in enumerate(var_dimensions):
            if var_dimension == 'time':
                dim_indexes.append(slice(None, None, None))
                continue
            dim_size = self._dimensions.get(var_dimension, -1)
            if dim_size < 0:
                raise ValueError(
                    f'Could not determine size of dimension {var_dimension}')
            data_offset = self._dimension_chunk_offsets.get(var_dimension, 0)
            start = data_offset + chunk_index[i + offset] * chunk_sizes[i]
            end = min(start + chunk_sizes[i], data_offset + dim_size)
            dim_indexes.append(slice(start, end))
        return tuple(dim_indexes)


def common_divisors(orig_number: int) -> List[int]:
    sqrt = math.sqrt(orig_number)
    factor = int(math.floor(sqrt))
    divisors = []
    if sqrt - factor < 1e-8:
        divisors.append(factor)
    while factor > 0:
        if orig_number % factor == 0:
            divisors.append(factor)
            divisors.append(int(orig_number / factor))
        factor -= 1
    return sorted(divisors, reverse=True)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any
from typing import Container
from typing import Dict
from typing import Iterator
from typing import Tuple
from typing import Union

import json

from xcube.core.store import DataStoreError
from xcube.core.store import DataTypeLike
from xcube.core.store import get_data_store_class
from xcube.util.jsonschema import JsonObjectSchema

CCI_ZARR_STORE_BUCKET_NAME = 'esacci'
CCI_ZARR_STORE_ENDPOINT = 'https://cci-ke-o.s3-ext.jc.rl.ac.uk:443/'
DATA_IDS_FILE_PATH = f'{CCI_ZARR_STORE_BUCKET_NAME}/data_ids.json'

CCI_ZARR_STORE_PARAMS = dict(
    root=CCI_ZARR_STORE_BUCKET_NAME,
    storage_options=dict(
        anon=True,
        client_kwargs=dict(
            endpoint_url=CCI_ZARR_STORE_ENDPOINT,
        )
    )
)

S3DataStore = get_data_store_class('s3')


class CciZarrDataStore(S3DataStore):

    def __init__(self):
        super().__init__(**CCI_ZARR_STORE_PARAMS)

    @classmethod
    def get_data_store_params_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(additional_properties=False)

    def get_data_ids(self,
                     data_type: DataTypeLike = None,
                     include_attrs: Container[str] | bool = False,) -> \
            Union[list[str], list[tuple[str, dict[str, Any]]]]:
        # TODO: do not ignore names in include_attrs
        if self.fs.exists(DATA_IDS_FILE_PATH):
            return_tuples = include_attrs and include_attrs is not None
            with self.fs.open(DATA_IDS_FILE_PATH) as f:
                ids = json.load(f)
                for id in ids:
                    yield (id, {}) if return_tuples else id
        else:
            yield from super().get_data_ids(
                data_type=data_type,
                include_attrs=include_attrs
            )

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def get_data_writer_ids(self, data_type: DataTypeLike = None) -> \
            Tuple[str, ...]:
        return ()

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def get_write_data_params_schema(self, **kwargs) -> \
            JsonObjectSchema:
        return JsonObjectSchema(additional_properties=False)

    def write_data(self, *args, **kwargs) -> str:
        raise DataStoreError('The CciZarrDataStore is read-only.')

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def get_delete_data_params_schema(self, **kwargs) -> \
            JsonObjectSchema:
        return JsonObjectSchema(additional_properties=False)

    def delete_data(self, *args):
        raise DataStoreError('The CciZarrDataStore is read-only.')



# The MIT License (MIT)
# Copyright (c) 2020 by the ESA CCI Toolbox development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import logging
import os
import os.path
import shutil
import time
import traceback
from typing import List, Optional

import click
import xarray as xr
import zarr.storage

from xcube.core.normalize import cubify_dataset
from xcube_cci.cciodp import CciOdp
from xcube_cci.chunkstore import CciChunkStore

DEFAULT_OUTPUT = 'open-report'


@click.command()
@click.option('--output', 'output_dir',
              metavar='OUTPUT_DIR',
              default=DEFAULT_OUTPUT,
              help=f'Output directory. Defaults to "{DEFAULT_OUTPUT}".')
@click.option('--force',
              is_flag=True,
              help=f'Force deletion of an existing OUTPUT_DIR.')
@click.option('--cache', 'cache_size',
              metavar='CACHE_SIZE',
              help=f'Cache size, e.g. "100M" or "2G". If given, an in-memory LRU cache will be used.')
@click.option('--observe',
              is_flag=True,
              help=f'Dump data requests. '
                   f'Note, turning this flag on will slightly increase measured durations.')
@click.argument('dataset_id', nargs=-1, required=False)
def gen_report(output_dir: str,
               force: bool,
               cache_size: Optional[str],
               observe: bool,
               dataset_id: List[str]):
    """
    Opens CCI ODP datasets and generates a report in OUTPUT_DIR from opening a dataset.

    If DATASET_ID are omitted, all ODP datasets are opened. Otherwise, only the given
    datasets will be opened. A DATASET_ID may be just a part of the actual dataset
    identifier to be selected, e.g. "CLOUD" or "5-day".

    For each selected dataset, the following reports are generated after opening it:

    \b
    * ${OUTPUT_DIR}/SUCCESS@${DATASET_ID}.json - on success only;
    * ${OUTPUT_DIR}/ERROR@${DATASET_ID}.json - on error only;
    * ${OUTPUT_DIR}/ERROR@${DATASET_ID}.txt - on error only, contains exception traceback.

    If the "--observe" flag is used, an additional report is generated comprising each
    individual data request made:

    * ${OUTPUT_DIR}/OBSERVED@${DATASET_ID}.txt - in any case.

    The program will exit with an error if OUTPUT_DIR exists. To force deletion of
    an existing OUTPUT_DIR, use the "--force" flag.

    """
    logging.basicConfig(format='%(asctime)s %(levelname)s: %(message)s',
                        level=logging.DEBUG,
                        datefmt='%Y-%m-%d %H:%M:%S')
    if os.path.isdir(output_dir):
        if force:
            shutil.rmtree(output_dir)
        else:
            raise click.ClickException(f'Output directory "{output_dir}" already exists.')
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    cache_size = parse_cache_size(cache_size)

    odp = CciOdp()

    selected_ds_ids = get_selected_dataset_ids(odp, dataset_id)

    logging.info(f'Running tests for {len(selected_ds_ids)} datasets...')

    id_to_ts = dict()

    for i, ds_id in enumerate(selected_ds_ids):
        data_type = ''
        observer, obs_fp = new_observer(output_dir, ds_id) if observe else (None, None)

        logging.info(f'Attempting to open #{i+1} of {len(selected_ds_ids)}: {ds_id}')

        t0 = time.perf_counter()

        try:
            store = CciChunkStore(odp, ds_id, observer=observer, trace_store_calls=True)
        except Exception as e:
            report_error(output_dir, ds_id, t0, 'CciChunkStore()', e, data_type)
            continue

        if cache_size > 0:
            store = zarr.storage.LRUStoreCache(store, max_size=cache_size)

        try:
            logging.info('Attempting to open zarr ...')
            ds = xr.open_zarr(store)
            data_type = 'dataset'
            report_success(output_dir, ds_id, t0, ds, data_type)
        except Exception as e:
            report_error(output_dir, ds_id, t0, 'xr.open_zarr()', e, data_type)

        id_to_ts[ds_id] = data_type

        if obs_fp is not None:
            obs_fp.close()

    report_type_specifiers(output_dir, ts_dict=id_to_ts)

    logging.info(f'Tests completed.')


def get_selected_dataset_ids(odp, ds_id_parts):
    all_ds_ids = odp.dataset_names
    if ds_id_parts:
        ds_id_parts_lc = [ds_id_part.lower() for ds_id_part in ds_id_parts]
        selected_ds_ids = set()
        for ds_id in all_ds_ids:
            ds_id_lc = ds_id.lower()
            if any(map(lambda ds_id_part: ds_id_part in ds_id_lc, ds_id_parts_lc)):
                selected_ds_ids.add(ds_id)
    else:
        selected_ds_ids = all_ds_ids
    selected_ds_ids = sorted(selected_ds_ids)
    return selected_ds_ids


def new_observer(output_dir, ds_id):
    path = os.path.join(output_dir, f'OBSERVED@{ds_id}.txt')
    fp = open(path, 'a')

    def observer(var_name: str = None, chunk_index=None, time_range=None, duration=None, exception=None):
        tr = f'{time_range[0]}/{time_range[1]}'
        data = f'{var_name};{chunk_index};{tr};{format_millis(duration)}'
        status = f'{type(exception).__name__}({exception})' if exception else 'OK'
        fp.write(f'{status};{data}\n')
        fp.flush()

    return observer, fp


def report_success(output_dir: str, ds_id: str, t0: float, ds: xr.Dataset, type_specifier: str):
    obj = ds_to_dict(ds_id, time.perf_counter() - t0, ds, type_specifier)
    path = os.path.join(output_dir, f'SUCCESS@{ds_id}.json')
    try:
        with open(path, 'w') as fp:
            json.dump(obj, fp, indent=2)
        logging.info(f'{ds_id} took {format_millis(obj["duration"])} seconds')
    except Exception as e:
        if os.path.isfile(path):
            os.remove(path)
        report_error(output_dir, ds_id, time.perf_counter() - t0, 'to_json()', e)


def report_error(output_dir: str, ds_id: str, t0: float, stage: str, e: Exception, type_specifier: str):
    obj = error_to_dict(ds_id, stage, time.perf_counter() - t0, e, type_specifier)
    json_path = os.path.join(output_dir, f'ERROR@{ds_id}.json')
    text_path = os.path.join(output_dir, f'ERROR@{ds_id}.txt')
    with open(json_path, 'w') as fp:
        json.dump(obj, fp, indent=2)
    with open(text_path, 'w') as fp:
        traceback.print_exc(file=fp)
    logging.error(f'{stage}: {ds_id} took {format_millis(obj["duration"])}: {e}')


def report_type_specifiers(output_dir: str, ts_dict: dict):
    json_path = os.path.join(output_dir, f'type_specifiers.json')
    with open(json_path, 'w') as fp:
        json.dump(ts_dict, fp, indent=2)


def ds_to_dict(ds_id: str, duration: float, ds: xr.Dataset, type_specifier: str) -> dict:
    return dict(ds_id=ds_id,
                status='ok',
                type_specifier=type_specifier,
                duration=duration,
                dataset=dict(sizes=dict(ds.sizes),
                             coord_vars=vars_to_list(ds.coords),
                             data_vars=vars_to_list(ds.data_vars),
                             attrs=dict(ds.attrs)))


def error_to_dict(ds_id: str, stage: str, duration: float, e: Exception, type_specifier: str) -> dict:
    return dict(ds_id=ds_id,
                typeSpecifier=type_specifier,
                status='error',
                duration=duration,
                error=dict(type=str(type(e)),
                           stage=stage,
                           message=f'{e}'))


def vars_to_list(vars) -> list:
    return [var_to_dict(k, v) for k, v in vars.items()]


def var_to_dict(name: str, var: xr.DataArray) -> dict:
    return dict(name=name,
                dtype=str(var.dtype),
                dims=list(var.dims),
                attrs=dict(var.attrs))


def parse_cache_size(cache_size: Optional[str]) -> int:
    if not cache_size:
        return 0
    cache_size_lc = cache_size.lower()

    def _parse_cache_size():
        for suffix, factor in (
                ('kib', 1024), ('mib', 1024 ** 2), ('gib', 1024 ** 3), ('tib', 1024 ** 4),
                ('b', 1), ('k', 1000), ('m', 1000 ** 2), ('g', 1000 ** 3), ('t', 1000 ** 4)):
            if cache_size_lc.endswith(suffix):
                return int(cache_size[0:-len(suffix)]) * factor
        return int(cache_size)

    try:
        size = _parse_cache_size()
        if size < 0:
            raise ValueError()
        return size
    except ValueError:
        raise click.ClickException(f'Invalid cache size: "{cache_size}"')


def format_millis(time: float) -> str:
    return f'{round(time * 1000)} ms'


if __name__ == '__main__':
    gen_report()



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from xcube.constants import EXTENSION_POINT_DATA_OPENERS
from xcube.constants import EXTENSION_POINT_DATA_STORES
from xcube.util import extension

from xcube_cci.constants import CCI_DATA_STORE_ID
from xcube_cci.constants import CDC_DATASET_OPENER_ID
from xcube_cci.constants import CDC_SHORT_DATA_STORE_ID
from xcube_cci.constants import CDC_LONG_DATA_STORE_ID
from xcube_cci.constants import DATASET_OPENER_ID
from xcube_cci.constants import DATA_STORE_ID
from xcube_cci.constants import KC_CCI_DATA_STORE_ID
from xcube_cci.constants import KC_CDC_DATA_STORE_ID
from xcube_cci.constants import KC_DATA_STORE_ID
from xcube_cci.constants import ZARR_CCI_DATA_STORE_ID
from xcube_cci.constants import ZARR_CDC_DATA_STORE_ID
from xcube_cci.constants import ZARR_DATA_STORE_ID


def init_plugin(ext_registry: extension.ExtensionRegistry):
    """xcube CCI ODP extensions"""
    data_completeness_content = \
        'This data store currently provides **only a subset of all ' \
        'datasets** provided by the "ESA CCI Climate Data Centre (CDC), ' \
        'namely gridded datasets originally stored in NetCDF format.\n' \
        'In upcoming versions, the store will also allow for browsing and ' \
        'accessing the remaining CDC datasets. This includes gridded data in ' \
        'TIFF format and also vector data using ESRI Shapefile format.'
    ext_registry.add_extension(
        loader=extension.import_component(
            'xcube_cci.dataaccess:CciOdpDataStore'),
        point=EXTENSION_POINT_DATA_STORES,
        name=CCI_DATA_STORE_ID,
        description='ESA Open Data Portal',
        data_store_notices=[dict(id='dataCompleteness',
                                 title='Data Completeness',
                                 content=data_completeness_content,
                                 intent='warning',
                                 icon='warning-sign')]
    )
    ext_registry.add_extension(
        loader=extension.import_component(
            'xcube_cci.dataaccess:CciOdpDataStore'),
        point=EXTENSION_POINT_DATA_STORES,
        name=CDC_SHORT_DATA_STORE_ID,
        description='ESA Open Data Portal',
        data_store_notices=[dict(id='dataCompleteness',
                                 title='Data Completeness',
                                 content=data_completeness_content,
                                 intent='warning',
                                 icon='warning-sign')]
    )
    ext_registry.add_extension(
        loader=extension.import_component(
            'xcube_cci.dataaccess:CciOdpDataStore'),
        point=EXTENSION_POINT_DATA_STORES,
        name=CDC_LONG_DATA_STORE_ID,
        description='ESA Open Data Portal',
        data_store_notices=[dict(id='dataCompleteness',
                                 title='Data Completeness',
                                 content=data_completeness_content,
                                 intent='warning',
                                 icon='warning-sign')]
    )
    ext_registry.add_extension(
        loader=extension.import_component(
            'xcube_cci.kcaccess:CciKerchunkDataStore'),
        point=EXTENSION_POINT_DATA_STORES,
        name=KC_CCI_DATA_STORE_ID,
        description='xarray.Dataset in Kerchunk references format'
                    ' from ESA CCI Object Storage'
    )
    ext_registry.add_extension(
        loader=extension.import_component(
            'xcube_cci.kcaccess:CciKerchunkDataStore'),
        point=EXTENSION_POINT_DATA_STORES,
        name=KC_CDC_DATA_STORE_ID,
        description='xarray.Dataset in Kerchunk references format'
                    ' from ESA CCI Object Storage'
    )
    ext_registry.add_extension(
        loader=extension.import_component(
            'xcube_cci.kcaccess:CciKerchunkDataStore'),
        point=EXTENSION_POINT_DATA_STORES,
        name=KC_DATA_STORE_ID,
        description='xarray.Dataset in Kerchunk references format'
                    ' from ESA CCI Object Storage'
    )
    ext_registry.add_extension(
        loader=extension.import_component(
            'xcube_cci.zarraccess:CciZarrDataStore'),
        point=EXTENSION_POINT_DATA_STORES,
        name=ZARR_CCI_DATA_STORE_ID,
        description='xarray.Dataset in Zarr format from ESA CCI Object Storage'
    )
    ext_registry.add_extension(
        loader=extension.import_component(
            'xcube_cci.zarraccess:CciZarrDataStore'),
        point=EXTENSION_POINT_DATA_STORES,
        name=ZARR_CDC_DATA_STORE_ID,
        description='xarray.Dataset in Zarr format from ESA CCI Object Storage'
    )
    ext_registry.add_extension(
        loader=extension.import_component(
            'xcube_cci.zarraccess:CciZarrDataStore'),
        point=EXTENSION_POINT_DATA_STORES,
        name=ZARR_DATA_STORE_ID,
        description='xarray.Dataset in Zarr format'
                    ' from ESA CCI Object Storage'
    )
    ext_registry.add_extension(
        loader=extension.import_component(
            'xcube_cci.dataaccess:CciOdpDataStore'),
        point=EXTENSION_POINT_DATA_STORES,
        name=DATA_STORE_ID,
        description='ESA CCI Open Data Portal',
        data_store_notices=[dict(id='dataCompleteness',
                                 title='Data Completeness',
                                 content=data_completeness_content,
                                 intent='warning',
                                 icon='warning-sign')])
    ext_registry.add_extension(
        loader=extension.import_component(
            'xcube_cci.dataaccess:CciOdpDatasetOpener'),
        point=EXTENSION_POINT_DATA_OPENERS,
        name=DATASET_OPENER_ID,
        description='xarray.Dataset in Zarr format'
                    ' from ESA CCI Open Data Portal'
    )
    ext_registry.add_extension(
        loader=extension.import_component(
            'xcube_cci.dataaccess:CciOdpDatasetOpener'),
        point=EXTENSION_POINT_DATA_OPENERS,
        name=CDC_DATASET_OPENER_ID,
        description='xarray.Dataset in Zarr format'
                    ' from ESA CCI Open Data Portal'
    )



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from datetime import datetime
from dateutil.relativedelta import relativedelta
import pandas as pd
from typing import Any
from typing import List
from typing import Mapping
from typing import Tuple
from typing import Union

from xcube_cci.cciodp import CciOdp

from xcube_cci.constants import MONTHS
from .constants import TIMESTAMP_FORMAT


def extract_time_range_as_strings(
        time_range: Union[Tuple, List]
) -> (str, str):
    if isinstance(time_range, tuple):
        time_start, time_end = time_range
    else:
        time_start = time_range[0]
        time_end = time_range[1]
    return \
        extract_time_as_string(time_start), \
        extract_time_as_string(time_end)


def extract_time_as_string(
        time_value: Union[pd.Timestamp, str]
) -> str:
    if isinstance(time_value, str):
        time_value = pd.to_datetime(time_value, utc=True)
    return time_value.tz_localize(None).isoformat()


class TimeRangeGetter:

    def __init__(self, cci_cdc: CciOdp, metadata: dict):
        self._cci_cdc = cci_cdc
        self._metadata = metadata

    def get_time_ranges(self, dataset_id: str,
                        params: Mapping[str, Any]) -> List[Tuple]:
        start_time, end_time, iso_start_time, iso_end_time = \
            self._extract_time_range_as_datetime(
                params.get('time_range',
                           self.get_default_time_range(dataset_id)))
        ecv = dataset_id.split('.')[1]
        time_period = dataset_id.split('.')[2]
        if time_period == 'day':
            start_time = datetime(year=start_time.year, month=start_time.month,
                                  day=start_time.day)
            end_time = datetime(year=end_time.year, month=end_time.month,
                                day=end_time.day,
                                hour=23, minute=59, second=59)
            delta = relativedelta(days=1)
        elif time_period == 'month' or time_period == 'mon':
            start_time = datetime(year=start_time.year, month=start_time.month,
                                  day=1)
            end_time = datetime(year=end_time.year, month=end_time.month, day=1)
            delta = relativedelta(months=1)
            end_time += delta
        elif time_period == 'year' or time_period == 'yr' and ecv != "BIOMASS":
            start_time = datetime(year=start_time.year, month=1, day=1)
            end_time = datetime(year=end_time.year, month=12, day=31)
            delta = relativedelta(years=1)
            if dataset_id.endswith("yr"):
                num_years = int(dataset_id[-3:-2])
                delta_years = relativedelta(years=num_years) - relativedelta(days=1)
                request_time_ranges = []
                this = start_time
                after = this + delta_years
                while after <= end_time:
                    pd_this = pd.Timestamp(datetime.strftime(this, TIMESTAMP_FORMAT))
                    pd_next = pd.Timestamp(datetime.strftime(after, TIMESTAMP_FORMAT))
                    request_time_ranges.append((pd_this, pd_next))
                    this = this + delta
                    after = after + delta
                return request_time_ranges
        elif time_period == 'climatology':
            return [(i + 1, i + 1) for i, month in enumerate(MONTHS)]
        else:
            end_time = end_time.replace(hour=23, minute=59, second=59)
            end_time_str = datetime.strftime(end_time, TIMESTAMP_FORMAT)
            iso_end_time = extract_time_as_string(end_time_str)
            request_time_ranges = self._cci_cdc.get_time_ranges_from_data(
                dataset_id, iso_start_time, iso_end_time
            )
            return request_time_ranges
        drs_end = dataset_id.split('.')[9]
        if drs_end == "greenland_gmb_mass_trends":
            return self.extract_greenland_gmb_time_series(
                dataset_id, iso_start_time, end_time, delta)
        request_time_ranges = []
        this = start_time
        while this < end_time:
            after = this + delta
            pd_this = pd.Timestamp(datetime.strftime(this, TIMESTAMP_FORMAT))
            pd_next = pd.Timestamp(datetime.strftime(after, TIMESTAMP_FORMAT))
            request_time_ranges.append((pd_this, pd_next))
            this = after
        return request_time_ranges

    def extract_greenland_gmb_time_series(
            self, dataset_id: str, iso_start_time: str,
            end_time: datetime, delta: pd.Timedelta
    )  -> List[Tuple]:
        end_time = end_time.replace(hour=23, minute=59, second=59)
        end_time_str = datetime.strftime(end_time, TIMESTAMP_FORMAT)
        iso_end_time = extract_time_as_string(end_time_str)
        request_time_ranges = self._cci_cdc.get_time_ranges_from_data(
            dataset_id, iso_start_time, iso_end_time
        )
        start_time = request_time_ranges[0][0]
        end_time = request_time_ranges[-1][1]
        num_years = end_time.year - start_time.year + 1
        span_delta = relativedelta(
            years=num_years - self._metadata["time_chunking"]
        )
        adjusted_time_ranges = []
        span_start_time = start_time
        span_end_time = start_time + span_delta
        span_end_time = span_end_time.replace(hour=23, minute=59, second=59)
        while span_end_time <= end_time:
            pd_this = pd.Timestamp(datetime.strftime(
                span_start_time, TIMESTAMP_FORMAT)
            )
            pd_next = pd.Timestamp(datetime.strftime(
                span_end_time, TIMESTAMP_FORMAT)
            )
            adjusted_time_ranges.append((pd_this, pd_next))
            span_start_time += delta
            span_end_time += delta
        return adjusted_time_ranges

    @staticmethod
    def _extract_time_range_as_datetime(
            time_range: Union[Tuple, List]
    ) -> (datetime, datetime, str, str):
        iso_start_time, iso_end_time = extract_time_range_as_strings(
            time_range)
        start_time = datetime.strptime(iso_start_time, TIMESTAMP_FORMAT)
        end_time = datetime.strptime(iso_end_time, TIMESTAMP_FORMAT)
        return start_time, end_time, iso_start_time, iso_end_time

    def get_default_time_range(self, ds_id: str):
        temporal_start = self._metadata.get('temporal_coverage_start', None)
        temporal_end = self._metadata.get('temporal_coverage_end', None)
        if not temporal_start or not temporal_end:
            time_ranges = self._cci_cdc.get_time_ranges_from_data(ds_id)
            if not temporal_start:
                if len(time_ranges) == 0:
                    raise ValueError(
                        "Could not determine temporal start of dataset. "
                        "Please use 'time_range' parameter."
                    )
                temporal_start = time_ranges[0][0]
            if not temporal_end:
                if len(time_ranges) == 0:
                    raise ValueError(
                        "Could not determine temporal end of dataset. "
                        "Please use 'time_range' parameter."
                    )
                temporal_end = time_ranges[-1][1]
        return temporal_start, temporal_end



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import unittest

import xarray as xr

from xcube.core.store import DataStore
from xcube.core.store import MutableDataStore
from xcube.core.store import DatasetDescriptor
from xcube.core.store import DataType
from xcube_cci.kcaccess import CciKerchunkDataStore

# os.environ['XCUBE_CCI_DISABLE_WEB_TESTS'] = '0'


@unittest.skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
                 'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
class CciKerchunkDataStoreOpenTest(unittest.TestCase):
    store = CciKerchunkDataStore()

    def test_open_data(self):
        data_id = "ESACCI-SOILMOISTURE-L3S-SSMS-ACTIVE-19910805-20211231-fv07.1_kr1.0"
        dataset = self.store.open_data(data_id)
        self.assertIsInstance(dataset, xr.Dataset)
        self.assertEqual({'time': 11107, 'lat': 720, 'lon': 1440}, dataset.sizes)


class CciKerchunkDataStoreTest(unittest.TestCase):
    store = CciKerchunkDataStore()

    def test_has_store(self):
        self.assertIsInstance(self.store, DataStore)
        self.assertNotIsInstance(self.store, MutableDataStore)

    def test_get_data_store_params_schema(self):
        self.assertIsNotNone(self.store.get_data_store_params_schema())

    def test_list_data_ids(self):
        data_ids = self.store.list_data_ids()
        self.assertIn(
            "ESACCI-BIOMASS-L4-AGB-CHANGE-100m-2018-2017-fv4.0-kr1.1",
            data_ids
        )
        self.assertIn(
            "ESACCI-L3C_CLOUD-CLD_PRODUCTS-MERIS-AATSR_ENVISAT-200301-201112_fv2.0_kr1.0",
            data_ids
        )
        self.assertIn(
            "CCI_GMB_GIS_DTU_fv1.5_kr1.0",
            data_ids
        )

    def test_get_data_opener_ids(self):
        self.assertEqual(("dataset:zarr:reference",), self.store.get_data_opener_ids())

    def test_get_data_types(self):
        self.assertEqual(("dataset",), self.store.get_data_types())
        self.assertEqual(("dataset",),
                         self.store.get_data_types_for_data("sst-cube"))

    def test_has_data(self):
        store = self.store
        data_id = 'ESACCI-L4_FIRE-BA-MODIS-20010101-20200120-fv5.1-kr1.2'
        self.assertEqual(True, store.has_data(data_id))
        self.assertEqual(False, store.has_data("lst-cube"))

    def test_describe_data(self):
        data_id = 'ESACCI-L4_FIRE-BA-MODIS-20010101-20200120-fv5.1-kr1.2'
        descriptor = self.store.describe_data(data_id)
        self.assertIsInstance(descriptor, DatasetDescriptor)
        self.assertEqual(data_id, descriptor.data_id)
        self.assertIsInstance(descriptor.data_type, DataType)
        self.assertIs(xr.Dataset, descriptor.data_type.dtype)
        self.assertIsInstance(descriptor.bbox, tuple)
        self.assertIsNone(descriptor.spatial_res)  # ?
        self.assertIsInstance(descriptor.dims, dict)
        self.assertIsInstance(descriptor.coords, dict)
        self.assertIsInstance(descriptor.data_vars, dict)
        self.assertIsInstance(descriptor.attrs, dict)

    def test_get_search_params_schema(self):
        # We do not have search parameters yet
        self.assertEqual(
            {
                "type": "object",
                "properties": {}
            },
            self.store.get_search_params_schema().to_dict()
        )

    def test_search_data(self):
        search_results = list(self.store.search_data())
        self.assertEqual(164, len(search_results))
        for descriptor, data_id in zip(search_results, self.store.get_data_ids()):
            self.assertIsInstance(descriptor, DatasetDescriptor)
            self.assertEqual(data_id, descriptor.data_id)
            self.assertIsInstance(descriptor.data_type, DataType)
            self.assertIs(xr.Dataset, descriptor.data_type.dtype)
            self.assertIsInstance(descriptor.dims, dict)
            self.assertIsInstance(descriptor.data_vars, dict)
            self.assertIsInstance(descriptor.attrs, dict)



import numpy
import os
import pandas as pd
import unittest
import xarray as xr

from unittest import skipIf

from xcube_cci.cciodp import CciOdp
from xcube_cci.chunkstore import CciChunkStore

OC_ID = 'esacci.OC.5-days.L3S.K_490.multi-sensor.multi-platform.MERGED.6-0.' \
        'sinusoidal'
OZONE_ID = 'esacci.OZONE.mon.L3.NP.multi-sensor.multi-platform.MERGED.fv0002.r1'
SEAICE_ID = 'esacci.SEAICE.day.L4.SICONC.multi-sensor.multi-platform.' \
            'AMSR_50kmEASE2.2-1.NH'
SST_ID = 'esacci.SST.satellite-orbit-frequency.L3U.SSTskin.AATSR.Envisat.' \
         'AATSR.2-1.r1'


class CciChunkStoreTest(unittest.TestCase):

    @staticmethod
    def _get_test_store():
        cci_cdc = CciOdp()
        dataset_id = OZONE_ID
        time_range = (pd.to_datetime('2010-02-10', utc=True),
                      pd.to_datetime('2010-05-20', utc=True))
        cube_params = dict(
            time_range=time_range,
            variable_names=['O3_vmr']
        )
        return CciChunkStore(cci_cdc, dataset_id, cube_params)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_unconstrained_chunk_store(self):
        cci_cdc = CciOdp()
        store = CciChunkStore(cci_cdc, OZONE_ID, cube_params=None)
        self.assertIsNotNone(store)
        time_ranges = store._time_ranges
        self.assertEqual(144, len(time_ranges))
        self.assertEqual(pd.Timestamp('1997-01-01T00:00:00'), time_ranges[0][0])
        self.assertEqual(
            {'surface_pressure', 'O3e_ndens', 'O3e_du', 'O3e_vmr', 'O3e_du_tot',
             'O3_du_tot', 'O3_ndens', 'O3_du', 'O3_vmr'},
            set(store._variable_names))

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_chunk_store_with_region_constraint(self):
        cci_cdc = CciOdp()
        cube_params = dict(bbox=[-10, 5, 0, 10])
        store = CciChunkStore(cci_cdc, OZONE_ID, cube_params=cube_params)
        self.assertIsNotNone(store)

        ds = xr.open_zarr(store)

        self.assertEqual(10, ds.lon.size)
        self.assertEqual(10, ds.lon.chunk_sizes)
        self.assertListEqual(
            [-9.5, -8.5, -7.5, -6.5, -5.5, -4.5, -3.5, -2.5, -1.5, -0.5],
            list(ds.lon.values))

        self.assertEqual(5, ds.lat.size)
        self.assertEqual(5, ds.lat.chunk_sizes)
        self.assertListEqual([5.5, 6.5, 7.5, 8.5, 9.5], list(ds.lat.values))

        self.assertEqual(['time', 'layers', 'lat', 'lon'],
                         ds.O3_du.dimensions)
        self.assertEqual((144, 16, 5, 10), ds.O3_du.shape)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_time_ranges(self):
        store = self._get_test_store()
        time_range = (pd.to_datetime('2010-02-10', utc=True),
                      pd.to_datetime('2010-05-20', utc=True))
        cube_params = dict(time_range=time_range)
        time_ranges = store.get_time_ranges(OZONE_ID, cube_params)
        self.assertEqual([('2010-02-01T00:00:00', '2010-03-01T00:00:00'),
                          ('2010-03-01T00:00:00', '2010-04-01T00:00:00'),
                          ('2010-04-01T00:00:00', '2010-05-01T00:00:00'),
                          ('2010-05-01T00:00:00', '2010-06-01T00:00:00')],
                         [(tr[0].isoformat(), tr[1].isoformat())
                          for tr in time_ranges])
        # get_time_range test for satellite-orbit-frequency data
        time_range = (pd.to_datetime('2002-07-24', utc=True),
                      pd.to_datetime('2002-07-24', utc=True))
        cube_params = dict(time_range=time_range)
        time_ranges = store.get_time_ranges(SST_ID, cube_params)
        self.assertEqual([('2002-07-24T12:33:21', '2002-07-24T14:13:57'),
                          ('2002-07-24T14:13:57', '2002-07-24T15:54:33'),
                          ('2002-07-24T15:54:33', '2002-07-24T17:35:09'),
                          ('2002-07-24T17:35:09', '2002-07-24T19:15:45'),
                          ('2002-07-24T19:15:45', '2002-07-24T20:56:21'),
                          ('2002-07-24T20:56:21', '2002-07-24T22:36:57'),
                          ('2002-07-24T22:36:57', '2002-07-25T00:17:33')],
                         [(tr[0].isoformat(), tr[1].isoformat())
                          for tr in time_ranges])
        # get_time_range test for data with days-period
        time_range = (pd.to_datetime('2002-07-04', utc=True),
                      pd.to_datetime('2002-07-27', utc=True))
        cube_params = dict(time_range=time_range)
        time_ranges = store.get_time_ranges(OC_ID, cube_params)
        self.assertEqual([('2002-06-30T00:00:00', '2002-07-04T23:59:00'),
                          ('2002-07-05T00:00:00', '2002-07-09T23:59:00'),
                          ('2002-07-10T00:00:00', '2002-07-14T23:59:00'),
                          ('2002-07-15T00:00:00', '2002-07-19T23:59:00'),
                          ('2002-07-20T00:00:00', '2002-07-24T23:59:00'),
                          ('2002-07-25T00:00:00', '2002-07-29T23:59:00')],
                         [(tr[0].isoformat(), tr[1].isoformat())
                          for tr in time_ranges])
        # get_time_range test for data with day-period
        time_range = (pd.to_datetime('2002-07-04', utc=True),
                      pd.to_datetime('2002-07-09', utc=True))
        cube_params = dict(time_range=time_range)
        time_ranges = store.get_time_ranges(SEAICE_ID, cube_params)
        self.assertEqual([('2002-07-04T00:00:00', '2002-07-05T00:00:00'),
                          ('2002-07-05T00:00:00', '2002-07-06T00:00:00'),
                          ('2002-07-06T00:00:00', '2002-07-07T00:00:00'),
                          ('2002-07-07T00:00:00', '2002-07-08T00:00:00'),
                          ('2002-07-08T00:00:00', '2002-07-09T00:00:00'),
                          ('2002-07-09T00:00:00', '2002-07-10T00:00:00')],
                         [(tr[0].isoformat(), tr[1].isoformat())
                          for tr in time_ranges])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_dimension_indexes_for_chunk(self):
        store = self._get_test_store()
        dim_indexes = store._get_dimension_indexes_for_chunk(
            'O3_vmr', (5, 0, 0, 0)
        )
        self.assertIsNotNone(dim_indexes)
        self.assertEqual(slice(None, None, None), dim_indexes[0])
        self.assertEqual(slice(0, 17), dim_indexes[1])
        self.assertEqual(slice(0, 180), dim_indexes[2])
        self.assertEqual(slice(0, 360), dim_indexes[3])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_encoding(self):
        store = self._get_test_store()
        encoding_dict = store.get_encoding('surface_pressure')
        self.assertTrue('fill_value' in encoding_dict)
        self.assertTrue('dtype' in encoding_dict)
        self.assertFalse('compressor' in encoding_dict)
        self.assertFalse('order' in encoding_dict)
        self.assertTrue(numpy.isnan(encoding_dict['fill_value']))
        self.assertEqual('float32', encoding_dict['dtype'])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_attrs(self):
        store = self._get_test_store()
        attrs = store.get_attrs('surface_pressure')
        self.assertTrue('standard_name' in attrs)
        self.assertTrue('long_name' in attrs)
        self.assertTrue('units' in attrs)
        self.assertTrue('fill_value' in attrs)
        self.assertTrue('chunk_sizes' in attrs)
        self.assertTrue('data_type' in attrs)
        self.assertTrue('dimensions' in attrs)
        self.assertEqual('surface_air_pressure', attrs['standard_name'])
        self.assertEqual('Pressure at the bottom of the atmosphere.',
                         attrs['long_name'])
        self.assertEqual('hPa', attrs['units'])
        self.assertTrue(numpy.isnan(attrs['fill_value']))
        self.assertEqual([1, 180, 360], attrs['chunk_sizes'])
        self.assertEqual('float32', attrs['data_type'])
        self.assertEqual(['time', 'lat', 'lon'], attrs['dimensions'])

    def test_adjust_chunk_sizes(self):
        chunk_sizes = [1, 128, 128]
        chunk_sizes = CciChunkStore._adjust_chunk_sizes(
            chunk_sizes, [2024, 4096, 4096], 0
        )
        self.assertEqual([1, 2048, 2048], chunk_sizes)

        chunk_sizes = [128, 128, 1]
        chunk_sizes = CciChunkStore._adjust_chunk_sizes(
            chunk_sizes, [4096, 4096, 2048], 2
        )
        self.assertEqual([2048, 2048, 1], chunk_sizes)

        chunk_sizes = [1, 128, 128]
        chunk_sizes = CciChunkStore._adjust_chunk_sizes(
            chunk_sizes, [2024, 128, 2048], 0
        )
        self.assertEqual([1, 128, 2048], chunk_sizes)

        chunk_sizes = [1, 64, 128, 32]
        chunk_sizes = CciChunkStore._adjust_chunk_sizes(
            chunk_sizes, [2048, 1024, 2048, 1024], 0
        )
        self.assertEqual([1, 64, 128, 32], chunk_sizes)

        chunk_sizes = [1, 90, 180]
        chunk_sizes = CciChunkStore._adjust_chunk_sizes(
            chunk_sizes, [1, 3600, 7200], 0
        )
        self.assertEqual([1, 1800, 1800], chunk_sizes)
        chunk_sizes = [1, 2048, 2048]
        chunk_sizes = CciChunkStore._adjust_chunk_sizes(
            chunk_sizes, [1, 64800, 129600], 0
        )
        self.assertEqual([1, 2025, 2025], chunk_sizes)

        chunk_sizes = [1, 4096, 4096]
        chunk_sizes = CciChunkStore._adjust_chunk_sizes(
            chunk_sizes, [1, 64800, 129600], 0)
        self.assertEqual([1, 2025, 2025], chunk_sizes
                         )

        chunk_sizes = [1]
        chunk_sizes = CciChunkStore._adjust_chunk_sizes(chunk_sizes, [1], -1)
        self.assertEqual([1], chunk_sizes)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import unittest

from xcube.core.store import DataStoreError
from xcube.util.jsonschema import JsonObjectSchema
from xcube_cci.zarraccess import CciZarrDataStore


@unittest.skipIf(os.environ.get('XCUBE_DISABLE_WEB_TESTS', None) == '1',
                 'XCUBE_DISABLE_WEB_TESTS = 1')
class CciZarrDataStoreTest(unittest.TestCase):

    def setUp(self) -> None:
        self.store = CciZarrDataStore()

    def test_get_data_store_params_schema(self):
        data_store_params_schema = self.store.get_data_store_params_schema()
        self.assertIsNotNone(data_store_params_schema)
        self.assertEqual(JsonObjectSchema(additional_properties=False).to_dict(),
                         data_store_params_schema.to_dict())

    def test_get_data_types(self):
        self.assertEqual({'mldataset', 'geodataframe', 'dataset'},
                         set(self.store.get_data_types()))

    def test_get_data_opener_ids(self):
        self.assertEqual({'dataset:geotiff:s3',
                          'dataset:netcdf:s3',
                          'dataset:zarr:s3',
                          'dataset:levels:s3',
                          'mldataset:geotiff:s3',
                          'mldataset:levels:s3',
                          'geodataframe:shapefile:s3',
                          'geodataframe:geojson:s3'},
                         set(self.store.get_data_opener_ids()))
        self.assertEqual({'dataset:geotiff:s3',
                          'dataset:netcdf:s3',
                          'dataset:zarr:s3',
                          'dataset:levels:s3'},
                         set(self.store.get_data_opener_ids(
                             data_type='dataset'
                         )))
        self.assertEqual({'dataset:geotiff:s3',
                          'dataset:netcdf:s3',
                          'dataset:zarr:s3',
                          'dataset:levels:s3',
                          'mldataset:geotiff:s3',
                          'mldataset:levels:s3',
                          'geodataframe:shapefile:s3',
                          'geodataframe:geojson:s3'},
                         set(self.store.get_data_opener_ids(
                             data_type='*'
                         )))
        self.assertEqual((), self.store.get_data_opener_ids(data_type=int))

    def test_get_open_data_params_schema(self):
        schema = self.store.get_open_data_params_schema()
        self.assertEqual(
            {'log_access',
             'cache_size',
             'chunks',
             'consolidated',
             'data_type',
             'decode_cf',
             'decode_coords',
             'decode_times',
             'drop_variables',
             'group',
             'mask_and_scale'},
            set(schema.properties.keys())
        )
        self.assertEqual([], schema.required)

    def test_get_data_writer_ids(self):
        self.assertEqual((), self.store.get_data_writer_ids())

    def test_get_write_data_params_schema(self):
        self.assertEqual(JsonObjectSchema(additional_properties=False).to_dict(),
                         self.store.get_write_data_params_schema().to_dict())

    def test_write_data(self):
        import xarray as xr
        data = xr.Dataset()
        with self.assertRaises(DataStoreError) as dse:
            self.store.write_data(data)
        self.assertEqual('The CciZarrDataStore is read-only.',
                         f'{dse.exception}')

    def test_delete_data(self):
        with self.assertRaises(DataStoreError) as dse:
            self.store.delete_data('some_data_id')
        self.assertEqual('The CciZarrDataStore is read-only.',
                         f'{dse.exception}')



#!/usr/bin/env python3

# The MIT License (MIT)
# Copyright (c) 2021 by Brockmann Consult GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from setuptools import setup, find_packages

requirements = [
    # Sync with ./environment.yml.
]

packages = find_packages(exclude=["test", "test.*"])

# Same effect as "from xcube_cci import version", but avoids importing xcube_cci:
version = None
with open('xcube_cci/version.py') as f:
    exec(f.read())

setup(
    name="xcube_cci",
    version=version,
    description='An xcube plugin that allows generating data cubes from the ESA CCI Open Data Portal',
    license='MIT',
    author='xcube Development Team',
    packages=packages,
    include_package_data=True,
    data_files=[('xcube_cci', [
        'xcube_cci/data/excluded_data_sources',
        'xcube_cci/data/stores.yml',
        'xcube_cci/data/kc_refs.json',
        'xcube_cci/data/dataset_states.json']
                 )],
    install_requires=requirements,
)



from unittest import TestCase
from xcube_cci.timeutil import get_timestrings_from_string


class TimeUtilTest(TestCase):

    def test_get_timestrings_from_string(self):
        timestring_1, timestring_2 = get_timestrings_from_string(
            '20020401-20020406-ESACCI-L3C_AEROSOL-AEX_ENVISAT-AERGOM_5days-fv2.19.nc'
        )
        self.assertEqual('2002-04-01T00:00:00', timestring_1)
        self.assertEqual('2002-04-06T00:00:00', timestring_2)

        timestring_1, timestring_2 = get_timestrings_from_string(
            '20020401-ESACCI-L3C_AEROSOL-AEX-GOMOS_ENVISAT-AERGOM_5days-fv2.19.nc'
        )
        self.assertEqual('2002-04-01T00:00:00', timestring_1)
        self.assertIsNone(timestring_2)

    def test_get_timestrings_from_string_months(self):
        timestring_1, timestring_2 = get_timestrings_from_string(
            'ESACCI-AEROSOL-L3-AAI-MSAAI-39Y1m_APRIL-fv1.7.nc'
        )
        self.assertEqual(4, timestring_1)
        self.assertIsNone(timestring_2)

        timestring_1, timestring_2 = get_timestrings_from_string(
            'ESACCI-AEROSOL-L3-AAI-MSAAI-39Y1m_OCTOBER-fv1.7.nc'
        )
        self.assertEqual(10, timestring_1)
        self.assertIsNone(timestring_2)

